@@ -38,33 +38,6 @@ def get_removed_attributes_and_base_net(net):
     return attributes, net
 
 
-def train_fold(net, trainloader, epochs, loss_function, optimizer, device, show_batch_interval=10_000, writer=None):
-    for epoch in range(epochs):
-        print(f"Epoch: {epoch + 1}")
-        running_loss = 0.0
-        epoch_steps = 0
-        for i, data in enumerate(trainloader, 0):
-            input, target = data
-            input, target = input.to(device), target.to(device)
-            optimizer.zero_grad()
-            output = net(input)
-            loss = loss_function(output, target)
-            loss.backward()
-            torch.nn.utils.clip_grad_norm_(net.parameters(), max_norm=1.0)
-            optimizer.step()
-            # print statistics
-            running_loss += loss.item()
-            epoch_steps += 1
-            if i % show_batch_interval == (show_batch_interval - 1):  # print every show_batch_interval mini-batches
-                print("Batch: %5d. Training Loss (running): %.3f" % (i + 1, running_loss / epoch_steps))
-                # Log the running loss averaged per batch
-                if writer is not None:
-                    writer.add_scalars(
-                        "Validation Loss", {"Validation": running_loss / epoch_steps}, epoch * len(trainloader) + i
-                    )
-                running_loss = 0.0
-
-
 def validate_fold_or_hold_out(net, valloader, loss_function, metric, device, task):
     val_loss = 0.0
     val_steps = 0
@@ -109,14 +82,25 @@ def validate_fold_or_hold_out(net, valloader, loss_function, metric, device, tas
 
 
 def evaluate_cv(
-    net, dataset, shuffle=False, num_workers=0, device=None, show_batch_interval=10_000, task=None, writer=None
+    net,
+    dataset,
+    shuffle=False,
+    loss_function=None,
+    num_workers=0,
+    device=None,
+    show_batch_interval=10_000,
+    metric=None,
+    path=None,
+    task=None,
+    writer=None,
+    writerId=None,
 ):
     lr_mult_instance = net.lr_mult
     epochs_instance = net.epochs
     batch_size_instance = net.batch_size
     k_folds_instance = net.k_folds
-    loss_function_instance = net.loss_function
     optimizer_instance = net.optimizer
+    patience_instance = net.patience
     sgd_momentum_instance = net.sgd_momentum
     removed_attributes, net = get_removed_attributes_and_base_net(net)
     metric_values = {}
@@ -135,7 +119,6 @@ def evaluate_cv(
             lr_mult=lr_mult_instance,
             sgd_momentum=sgd_momentum_instance,
         )
-        loss_function = loss_function_instance
         kfold = KFold(n_splits=k_folds_instance, shuffle=shuffle)
         for fold, (train_ids, val_ids) in enumerate(kfold.split(dataset)):
             print(f"Fold: {fold + 1}")
@@ -147,23 +130,56 @@ def evaluate_cv(
             valloader = torch.utils.data.DataLoader(
                 dataset, batch_size=batch_size_instance, sampler=val_subsampler, num_workers=num_workers
             )
+            # each fold starts with new weights:
             reset_weights(net)
-            # Train fold for several epochs:
-            train_fold(
-                net,
-                trainloader,
-                epochs_instance,
-                loss_function,
-                optimizer,
-                device,
-                show_batch_interval=show_batch_interval,
-                writer=writer,
-            )
-            # Validate fold: use only loss for tuning
-            metric_values[fold], loss_values[fold] = validate_fold_or_hold_out(
-                net, valloader, loss_function, device, task
-            )
+            # Early stopping parameters
+            best_val_loss = float("inf")
+            counter = 0
+            for epoch in range(epochs_instance):
+                print(f"Epoch: {epoch + 1}")
+                # training loss from one epoch:
+                training_loss = train_hold_out(
+                    net=net,
+                    trainloader=trainloader,
+                    batch_size=batch_size_instance,
+                    loss_function=loss_function,
+                    optimizer=optimizer,
+                    device=device,
+                    show_batch_interval=show_batch_interval,
+                    task=task,
+                    writer=writer,
+                )
+                # TODO: scheduler.step()
+                # Early stopping check. Calculate validation loss from one epoch:
+                metric_values[fold], loss_values[fold] = validate_fold_or_hold_out(
+                    net, valloader=valloader, loss_function=loss_function, metric=metric, device=device, task=task
+                )
+                # Log the running loss averaged per batch
+                metric_name = "Metric"
+                if metric is None:
+                    metric_name = type(metric).__name__
+                    print(f"{metric_name} value on hold-out data: {metric_values[fold]}")
+                if writer is not None:
+                    writer.add_scalars(
+                        "evaluate_cv fold:" + str(fold + 1) + ". Train & Val Loss and Val Metric" + writerId,
+                        {"Train loss": training_loss, "Val loss": loss_values[fold], metric_name: metric_values[fold]},
+                        epoch + 1,
+                    )
+                    writer.flush()
+                if loss_values[fold] < best_val_loss:
+                    best_val_loss = loss_values[fold]
+                    counter = 0
+                    # save model:
+                    if path is not None:
+                        torch.save(net.state_dict(), path)
+                else:
+                    counter += 1
+                    if counter >= patience_instance:
+                        print(f"Early stopping at epoch {epoch}")
+                        break
+        # TODO: Compute Metric on all folds and return average to spotPython
         df_eval = sum(loss_values.values()) / len(loss_values.values())
+        df_metrics = sum(metric_values.values()) / len(metric_values.values())
         df_preds = np.nan
     except Exception as err:
         print(f"Error in Net_Core. Call to evaluate_cv() failed. {err=}, {type(err)=}")
@@ -171,8 +187,16 @@ def evaluate_cv(
         df_preds = np.nan
     add_attributes(net, removed_attributes)
     if writer is not None:
+        metric_name = "Metric"
+        if metric is None:
+            metric_name = type(metric).__name__
+        writer.add_scalars(
+            "CV: Val Loss and Val Metric" + writerId,
+            {"CV-loss": df_eval, metric_name: df_metrics},
+            epoch + 1,
+        )
         writer.flush()
-    return df_eval, df_preds
+    return df_eval, df_preds, df_metrics
 
 
 def evaluate_hold_out(
